#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2012 Liricus SRL <info@liricus.com.ar>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Fra


#===============================================================================
# DOCS
#===============================================================================

"""Application routes that conform the base api.
All the api calls return a JSON object with information on the requested action.

    **JSON:**
        :action:
            The api function called.
        :code:
            HTML status code.
        :error:
            Boolean, ``false`` if nothing weird happened ``true`` otherwise.
        :traceback:
            Apperas when ``error`` is ``true`` and ``DEBUG`` is ``True`` in
            ``settings`` Python traceback most recent call.
        :msg:
            Appears when ``error`` is ``true``, it's the html error message.

"""


#===============================================================================
# IMPORTS
#===============================================================================

import string
import random
import datetime
import urllib

import peewee

import flask

from socialframe_server import core
from socialframe_server import models
from socialframe_server import settings
from socialframe_server import emailutil


#===============================================================================
# ROUTES
#===============================================================================

@core.api("/", cached=86400)
def index():
    """Returns a message if the server work

    **Logged in:** False

    **Method:** GET

    **Return**

        ::

            {"error": false,
             "action": "index",
             "code": 200,
             "msg": "its works"}

    """
    return {"msg": "its works"}


@core.api("/sizes", cached=86400)
def sizes():
    """Return al aivailable ``IMAGE_SIZES`` for images configured in enviroment
    ``settings.py``.

    **Logged in:** False

    **Method:** GET

    ***Return**

        ::

            {"error": false,
             "action": "sizes",
             "code": 200, "sizes": {"size0": <int>, ...}}

    """
    return {"sizes": settings.IMAGE_SIZES}


@core.api("/apihelp", cached=86400)
def apihelp():
    """Returns api documentation.

    **Logged in:** False

    ***Return**

        ::

            {"error": false,
             "action": "apihelp",
             "code": 200, "api": {"route0": doc0, ...}}

    """
    data = {}
    for route in core.APP.url_map.iter_rules():
        doc = getattr(core.APP.view_functions[route.endpoint], "__doc__", "")
        if doc:
            doc = doc.splitlines()[0]
        else:
            doc = "<NO-DOC>"
        data[core.absurl(route.rule)] = doc
    return {"api": data}


@core.api("/quick_register", cached=10, transaction=True, methods=['POST'])
def quick_register():
    """Register and activate a new user into the server.

    **Logged in:** False

    **Method:** POST

    **Params**
        :email: User email.
        :password: password.

        **Return**

            ::

                {"error": false,
                 "action": "quick_register",
                 "code": 200}

        Also an email is send to a user with a link to confirm activation.

    """
    user = models.User()
    user.email = urllib.unquote_plus(flask.request.form["email"])
    user.fname = user.email
    user.lname = user.email
    user.birth = datetime.datetime.now().date()
    user.set_password(urllib.unquote_plus(flask.request.form["password"]))
    user.set_status(user.STATUS_ACTIVE)
    user.save()

    data = user.email
    emailutil.send_mail(settings.MAIL_QUICK_CREATION, user,
                        flask.request.accept_languages.values(),
                        data)


@core.api("/register_user", cached=10, transaction=True, methods=['POST'])
def register_user():
    """Register a new user into the server.

    **Logged in:** False

    **Method:** POST

    **Params**
        :email: User email.
        :fname: First name of the user.
        :lname: Last name
        :birth: Date with format YYYYMMDD (19860808 for 1986-08-08) format.
        :sex: M or F.
        :password: password.

        **Return**

            ::

                {"error": false,
                 "action": "register_user",
                 "code": 200}

        Also an email is send to a user with a link to confirm registration.

    """
    user = models.User()
    user.email = urllib.unquote_plus(flask.request.form["email"])
    user.fname = urllib.unquote_plus(flask.request.form["fname"])
    user.lname = urllib.unquote_plus(flask.request.form["lname"])
    user.birth = core.parsedate(urllib.unquote_plus(flask.request.form["birth"])).date()
    user.sex = urllib.unquote_plus(flask.request.form["sex"].lower())
    user.set_password(urllib.unquote_plus(flask.request.form["password"]))
    user.save()
    profile = models.Profile.get(models.Profile.user == user)
    profile.operation = profile.OPERATION_CREATE_USER
    profile.save()
    data = core.absurl("confirm_registration", user.email,
                       profile.operation_hash)
    emailutil.send_mail(settings.MAIL_CREATION, user,
                        flask.request.accept_languages.values(),
                        data)


@core.api("/confirm_registration/<email:email>/<userhash>", cached=10)
def confirm_registration(email, userhash):
    """Activate a user by a given email and his activation hash.

    **Logged in:** False

    **Method:** GET

    **Params**
        :email:
            Email of the user to activate.
        :userhash:
            Hash sended as email to the user email

    **Return**

        ::

            {"action": "confirm_registration",
             "code": 200,
             "error": false}

    """
    user = models.User.get(models.User.email == email)
    if not user.active:
        profile = models.Profile.get(models.Profile.user == user)
        if profile.validate_hash(profile.OPERATION_CREATE_USER, userhash):
            user.set_status(user.STATUS_ACTIVE)
            user.last_know_lang = (flask.request.accept_languages.values()[0]
                                   if flask.request.accept_languages.values()
                                   else settings.DEFAULT_LANG)
            user.last_login = datetime.datetime.now()
            user.save()
        else:
            flask.abort(403)


@core.api("/user_data/<email:email>", cached=60, loggedin=True)
def user_data(email):
    """Retrieve all the data of a given user identified by his email.

    **Logged in:** True

    **Method:** GET

    **Params**
        :email:
            Email of the user to see the data.

    **Return**

        ::

            {"action": "user_data",
             "code": 200,
             "error": false,
             "user": {<USER-DATA>}}

    """
    user = models.User.get(models.User.email == email)
    return {"user": user.to_dict()}


@core.api("/reactivate_user/<email:email>/<userhash>", cached=10)
def reactivate_user(email, userhash):
    """This method reactivate a user previously suspended.

    **Logged in:** False

    **Method:** GET

    **Params**
        :email:
            Email of the user to activate.
        :userhash:
            Hash sended as email to the user email

    **Return**

        ::

            {"action": "reactivate_user",
             "code": 200,
             "error": false}

    """
    user = models.User.get(models.User.email == email)
    if not user.active:
        profile = models.Profile.get(models.Profile.user == user)
        if profile.validate_hash(models.Profile.OPERATION_REACTIVATE_USER, userhash):
            user.set_status(user.STATUS_ACTIVE)
            user.last_know_lang = (flask.request.accept_languages.values()[0]
                                   if flask.request.accept_languages.values()
                                   else settings.DEFAULT_LANG)
            user.last_login = datetime.datetime.now()
            user.save()
        else:
            flask.abort(403)


@core.api("/change_user", loggedin=True, methods=['POST'])
def change_user():
    """Change the user data.

    **Logged in:** True

    **Method:** POST

    **Params**
        :email: User email *(optional)*.
        :fname: First name of the user *(optional)*.
        :lname: Last name *(optional)*.
        :birth: Date with format YYYYMMDD (19860808 for 1986-08-08) format
                *(optional)*.
        :sex: ``M`` or ``F`` *(optional)*.
        :password: password *(optional)*..

    **Return**

        ::

            {"action": "change_user",
             "code": 200,
             "error": false}

    """
    user = core.APP_AUTH.get_logged_in_user()
    if user.is_superuser:
        return flask.abort(403)
    user.fname = urllib.unquote_plus(flask.request.form.get("fname", user.fname))
    user.lname = urllib.unquote_plus(flask.request.form.get("lname", user.lname))
    user.sex = urllib.unquote_plus(flask.request.form.get("sex", user.sex)).lower()
    if "birth" in flask.request.form:
        user.birth = core.parsedate(urllib.unquote_plus(flask.request.form["birth"])).date()
    if "password" in flask.request.form:
        user.set_password(urllib.unquote_plus(flask.request.form["password"]))
    user.save()


@core.api("/is_loggedin", cached=60)
def is_loggedin():
    """Checks for current user is logged in.

    The value of the key ``is_loggedin`` can be ``true`` or ``false``.

    **Logged in:** False

    **Method:** GET

    **Return**

        ::

            {"action": "is_loggedin",
             "code": 200,
             "is_loggedin": true,
             "error": false}

    """
    loggedin = bool(core.APP_AUTH.get_logged_in_user())
    return {"is_loggedin": loggedin}


@core.api("/login", cached=10, transaction=True, methods=["POST"])
def login():
    """User Login.

    **Logged in:** False

    **Method:** POST

    **Parameters**

        :email: User email.
        :password: User password.


    **Return**

        ::

            {"error": false,
             "action": "login",
             "code": 200}

    """
    email = urllib.unquote_plus(flask.request.form["email"])
    password = urllib.unquote_plus(flask.request.form["password"])
    user = models.User.get(models.User.email == email, models.User.active == True)
    if not user or not user.check_password(password):
        flask.abort(403)
    core.APP_AUTH.login_user(user)
    user.last_login = datetime.datetime.now()
    user.last_know_lang = (flask.request.accept_languages.values()[0]
                           if flask.request.accept_languages.values()
                           else settings.DEFAULT_LANG)
    user.save()


@core.api("/logout")
def logout():
    """Logout user. If no user it's logged throws ``Error 403``.

    **Logged in:** True

    **Method:** GET

    **Return**

        ::

            {"error": false,
             "action": "logout",
             "code": 200}

    """
    core.APP_AUTH.logout_user(core.APP_AUTH.get_logged_in_user())


@core.api("/reset_password", transaction=True, methods=["POST"])
def reset_password():
    """Method for send a mail to user for reset his password.

    **Logged in:** False

    **Method:** POST

    **Parameters**

        :email:
            User email.

    **Return**

        ::

            {"action": "reset_password",
             "code": 200,
             "error": false}

        Also an email is sent with the new password.

    """
    def gen_password(size=8):
        chars = string.letters + string.digits
        newpasswd = ""
        for _ in range(size):
            newpasswd += random.choice(chars)
        return newpasswd

    email = urllib.unquote_plus(flask.request.form["email"])
    user = models.User.get(models.User.email == email,
                           models.User.is_superuser == False)
    if not user:
        flask.abort(403)
    new_password = gen_password()
    user.set_password(new_password)
    user.save()

    emailutil.send_mail(settings.MAIL_RESET_PASSWORD, user,
                        flask.request.accept_languages.values(),
                        new_password)


@core.api("/delete_user", loggedin=True, cached=10,
          transaction=True, methods=["POST"])
def delete_user():
    """send a email to the user for delete his account.

    **Logged in:** True

    **Method:** POST

    **Params**
        :password: the actual logged in user

        **Return**

            {"action": "delete_user","code": 200,"error": false}

            Also an email is sent to confirm the action.

    """
    user = core.APP_AUTH.get_logged_in_user()
    password = urllib.unquote_plus(flask.request.form["password"])
    if not user.check_password(password) or user.is_superuser:
        flask.abort(403)
    profile = models.Profile.get(models.Profile.user == user)
    profile.operation = models.Profile.OPERATION_DELETE_USER
    profile.save()

    data = core.absurl("confirm_delete", user.email, profile.operation_hash)
    emailutil.send_mail(settings.MAIL_DELETE, user,
                        flask.request.accept_languages.values(),
                        data)


@core.api("/confirm_delete/<email:email>/<deletehash>", cached=60,
          loggedin=True, transaction=True)
def confirm_delete(email, deletehash):
    """Confirm and proced to delete the user account and bubbles.

    **Logged in:** True

    **Method:** GET

    **Params**

        :email:
            User email.
        :deletehash:
            Operation hash.

    **Return**

        ::

            {"action": "confirm_delete",
             "code": 200,
             "error": false}

    """
    user = core.APP_AUTH.get_logged_in_user()
    if user.email != email or user.is_superuser:
        flask.abort(403)
    profile = models.Profile.get(models.Profile.user == user)
    if profile.validate_hash(profile.OPERATION_DELETE_USER, deletehash):
        user.delete_instance(recursive=True)
        profile.delete_instance(recursive=True)
    else:
        flask.abort(403)


@core.api("/create_bubble", loggedin=True, cached=5,
          transaction=True, methods=["POST"])
def create_bubble():
    """Creates a new bubble to the current user.

    **Logged in:** True

    **Method:** POST

    **Params**

        :bubble_name: The name of the bubble to be modified.
        :desc: The description of the bubble.
        :type: The type of the bubble ["community"|"public"].
        :listed: If the bubble must be listed in the search and random
                 [true|false]
        :password: The password for acess the bubble phothos or "null"
        :tags: A coma separated list of tags for identify the bubble or
               "null".

    **Return**

        ::

            {"action": "create_bubble",
             "code": 200,
             "error": false}

    """
    user = core.APP_AUTH.get_logged_in_user()
    bubble = models.Bubble()
    bubble.user = user
    bubble.name = urllib.unquote_plus(flask.request.form["bubble_name"])
    bubble.desc = urllib.unquote_plus(flask.request.form["desc"])
    bubble.type = urllib.unquote_plus(flask.request.form["type"])
    if bubble.type == models.Bubble.ENTERPRISE_BUBBLE \
       and not user.is_enterprise:
               return flask.abort(403)
    if urllib.unquote_plus(flask.request.form["listed"]) == "true":
        bubble.listed = True
    bubble_password = urllib.unquote_plus(flask.request.form["password"])
    if  bubble_password != "null":
        bubble.set_password(bubble_password)
    bubble.save()
    tags = urllib.unquote_plus(flask.request.form["tags"])
    if tags != "null":
        tags = tags.split(",")
        bubble.set_tags(tags)


@core.api("/change_bubble", loggedin=True, transaction=True, methods=['POST'])
def change_bubble():
    """Changes the values of a bubble

    **Logged in:** True

    **Method:** POST

    **Params**

        :bubble_name: The name of the bubble to be modified.
        :new_bubble_name: The new name of the bubble *(optional)*..
        :desc: The description of the bubble *(optional)*..
        :type: The type of the bubble ["community"|"public"] *(optional)*..
        :listed: If the bubble must be listed in the search and random
                 [true|false] *(optional)*.
        :password: The password for acess the bubble phothos or "null"
                   *(optional)*.
        :tags: A coma separated list of tags for identify the bubble.
               *(optional)*.

    **Return**

        ::

            {"action": "change_bubble",
             "code": 200,
             "error": false}

    """
    user = core.APP_AUTH.get_logged_in_user()
    bubble_name = urllib.unquote_plus(flask.request.form["bubble_name"])
    bubble = models.Bubble.get(models.Bubble.name == bubble_name,
                               models.Bubble.user == user)
    if "new_bubble_name" in flask.request.form:
        bubble.name = urllib.unquote_plus(flask.request.form["new_bubble_name"])
    if "desc" in flask.request.form:
        bubble.desc = urllib.unquote_plus(flask.request.form["desc"])
    if "type" in flask.request.form:
        bubble.type = urllib.unquote_plus(flask.request.form["type"])
        if bubble.type == models.Bubble.ENTERPRISE_BUBBLE \
           and not user.is_enterprise:
               return flask.abort(403)
    if "listed" in flask.request.form:
        listed = urllib.unquote_plus(flask.request.form["listed"])
        bubble.listed = (listed == "true")
    if "password" in flask.request.form:
        password = urllib.unquote_plus(flask.request.form["password"])
        password = None if password == "null" else password
        bubble.set_password(password)
    if "tags" in flask.request.form:
        tags = urllib.unquote_plus(flask.request.form["tags"])
        tags = [] if tags == "null" else tags.split(",")
        bubble.set_tags(tags)
    bubble.save()


@core.api("/list_bubbles/<email:email>", cached=10)
def list_bubbles(email):
    """Returns only listed ``bubbles`` from  user ``email``.

    **Logged in:** True

    **Method:** GET

    **Params**
        :email: User email.

    **Return**

        ::

            {
              "action": "list_bubbles",
              "bubbles": {
                "test1": {
                  "has_password": true,
                  "listed": true,
                  "tags": []
                },
                "test3": {
                  "has_password": false,
                  "listed": true,
                  "tags": []
                }
              },
              "code": 200,
              "error": false
            }

    """
    data = {}
    query = models.Bubble.select().join(models.User)
    query = query.where(models.User.email == email,
                        models.Bubble.listed == True)
    for bubble in query:
        data[bubble.name] = bubble.to_dict()
    return {"bubbles": data}


@core.api("/my_bubbles", loggedin=True, cached=10)
def my_bubbles():
    """Returns all the current user bubbles.

    **Logged in:** True

    **Method:** GET

    **Return**

        ::

            {
              "action": "my_bubbles",
              "code": 200,
              "my_bubbles": {
                "test1": {
                  "has_password": true,
                  "listed": true,
                  "tags": []
                },
                "test3": {
                  "has_password": false,
                  "listed": true,
                  "tags": []
                },
                "test2": {
                  "has_password": true,
                  "listed": false,
                  "tags": []
                }
              },
              "error": false
            }

    """
    user = core.APP_AUTH.get_logged_in_user()
    data = {}
    for bubble in models.Bubble.select().where(models.Bubble.user == user):
        data[bubble.name] = bubble.to_dict()
    for bc in models.BubbleCommiter.select().where(models.BubbleCommiter.user == user):
        bubble = bc.bubble
        data[bubble.name] = bubble.to_dict()
    return {"bubbles": data}


@core.api("/validate_bubble_password/<email:email>/<bubble_name>/<null:password>",
           cached=10)
def validate_bubble_password(email, bubble_name, password):
    """Returns true or false if the password is valid for a given bubble.

    **Logged in:** False

    **Method:** GET

    **Params**
        :email: User email.
        :bubble_name: Name of a existing bubble of the given email.
        :password: The password of the bubble or "null".

    **Return**

        ::

            {"action": "validate_bubble_password",
             "code": 200,
             "valid_password": true.
             "error": false}

    """
    user = models.User.get(models.User.email == email)
    bubble_name = urllib.unquote_plus(bubble_name)
    bubble = models.Bubble.get(models.Bubble.user == user,
                               models.Bubble.name == bubble_name)
    bpass = bubble.check_password(password)
    return {"valid_password": bpass}


@core.api("/list_photos/<email:email>/<bubble_name>/<null:password>", cached=10)
def list_photos(email, bubble_name, password):
    """Returns a list of the user (``email``) photos of a given in the
    specified ``bubble_name``.

    **Logged in:** False

    **Params**

        :email:
            User email.
        :bubble_name:
            Bubble name.
        :password[string, null]:
            Password.

    **Return**

        ::

            {
              "action": "list_photos",
              "photos": {
                "a7e804d33aa585bmodels.30bec77896fa9512eea797.jpg": {
                  "sha": "f45cf037f11ca3ade43f59be3bb0c410b5b0fbf5"
                  "filename": "a7e804d33aa585bmodels.30bec77896fa9512eea797.jpg",
                  "status": <STATUS>,
                  "desc": "image_description"
                },
              },
              "error": false
                }

    """
    photos = {}
    user = models.User.get(models.User.email == email)
    bubble_name = urllib.unquote_plus(bubble_name)
    bubble = models.Bubble.get(models.Bubble.user == user,
                               models.Bubble.name == bubble_name)
    if bubble.has_password and not bubble.check_password(password):
        flask.abort(403)
    query = models.Image.select().where(models.Image.bubble == bubble)
    query = query.where(models.Image.status == models.Image.STATUS_APPROVED)
    for image in query:
        photos[image.name] = image.to_dict()
    return {"photos": photos}


@core.api("/list_photos_eb/<bubble_name>",
            cached=10, loggedin=True, enterprise=True)
def list_photos_eb(bubble_name):
    """Returns a list of all photos of a given enterprise bubble of the logged
    in user.

    **Logged in:** True

    **Parameters**

        :bubble_name:
            Bubble name.

    **Return**

        ::

            {
              "action": "list_photos",
              "photos": {
                "a7e804d33aa585bmodels.30bec77896fa9512eea797.jpg": {
                  "sha": "f45cf037f11ca3ade43f59be3bb0c410b5b0fbf5"
                  "filename": "a7e804d33aa585bmodels.30bec77896fa9512eea797.jpg",
                  "status": <STATUS>,
                  "desc": "image_description"
                },
              },
              "error": false
                }

    """
    user = core.APP_AUTH.get_logged_in_user()
    photos = {}
    bubble_name = urllib.unquote_plus(bubble_name)
    bubble = models.Bubble.get(models.Bubble.user == user,
                               models.Bubble.name == bubble_name,
                               models.Bubble.type == models.Bubble.ENTERPRISE_BUBBLE)
    query = models.Image.select().where(models.Image.bubble == bubble)
    for image in query:
        photos[image.name] = image.to_dict()
    return {"photos": photos}



@core.api("/delete_bubble", loggedin=True,
          transaction=True, cached=10, methods=["POST"])
def delete_bubble():
    """Deletes a bubble from the current user.

        **Parameters**
            :bubble_name:
                Name of bubble.

        **Return**

            {"action": "delete_bubble", "code": 200, "error": false}

    """
    user = core.APP_AUTH.get_logged_in_user()
    bubble_name = urllib.unquote_plus(flask.request.form["bubble_name"])
    bubble = models.Bubble.get(models.Bubble.name == bubble_name,
                               models.Bubble.user == user)
    bubble.delete_instance(recursive=True)


@core.api("/get_photo/<email:email>/<bubble_name>/<filename>/<size>/<null:password>",
          json=False, transaction=True, cached=120)
def get_photo(email, bubble_name, filename, size, password):
    """Looks for a photo (``filename``) in the specified ``bubble_name`` of the
    user ``email`` and returns the photo.
    If the bubble has a password and the one given does not match, ``Error 403``
    is raised.

        **Parameters**
            :email:
                User email.
            :bubble_name:
                Bubble name.
            :filename:
                Photo name.
            :password[string, null]:
                Password.

    """
    size = urllib.unquote_plus(size)
    if size not in settings.IMAGE_SIZES:
        flask.abort(400)
    bubble_name = urllib.unquote_plus(bubble_name)
    filename = urllib.unquote_plus(filename)
    user = models.User.get(models.User.email == email)
    bubble = models.Bubble.get(models.Bubble.user == user,
                               models.Bubble.name == bubble_name)
    if bubble.has_password and not bubble.check_password(password):
        flask.abort(403)

    image = models.Image.get(models.Image.bubble == bubble,
                             models.Image.name == filename,
                             models.Image.status == models.Image.STATUS_APPROVED)
    return flask.send_file(image.get_size(size).path, as_attachment=True)


@core.api("/get_photo_eb/<bubble_name>/<filename>/<size>",
           json=False, transaction=True, cached=120,
           loggedin=True, enterprise=True)
def get_photo_eb(bubble_name, filename, size):
    """Give any photo of enterprise bubble of the logged in user

    **Logged in:** False

    **Method:** GET

    **Params**
        :email:
            User email.
        :bubble_name:
            Bubble name.
        :filename:
            Photo name.
        :size:
            A registered sized

    """
    size = urllib.unquote_plus(size)
    if size not in settings.IMAGE_SIZES:
        flask.abort(400)
    user = core.APP_AUTH.get_logged_in_user()
    bubble_name = urllib.unquote_plus(bubble_name)
    filename = urllib.unquote_plus(filename)
    bubble = models.Bubble.get(models.Bubble.user == user,
                               models.Bubble.name == bubble_name)
    image = models.Image.get(models.Image.bubble == bubble,
                             models.Image.name == filename)
    return flask.send_file(image.get_size(size).path, as_attachment=True)


@core.api("/delete_photo", loggedin=True,
            transaction=True, cached=10, methods=["POST"])
def delete_photo():
    """Deletes a photo from a bubble

        **Parameters**
            :bubble_name:
                Name of bubble.
            :filename:
                Photo name.

    """
    user = core.APP_AUTH.get_logged_in_user()
    bubble_name = urllib.unquote_plus(flask.request.form["bubble_name"])
    filename = urllib.unquote_plus(flask.request.form["filename"])
    bubble = models.Bubble.get(models.Bubble.user == user,
                               models.Bubble.name == bubble_name)
    image = models.Image.get(models.Image.bubble == bubble,
                             models.Image.name == filename)
    image.delete_instance(True)


@core.api("/upload_photo", loggedin=True,
          transaction=True, methods=["POST"])
def upload_photo():
    """Uploads a photo to the specified bubble.

        **Parameters**
            :bubble_name:
                Name of the bubble to upload

    """
    user = core.APP_AUTH.get_logged_in_user()
    email = urllib.unquote_plus(flask.request.form["email"])
    bubble_name = urllib.unquote_plus(flask.request.form["bubble_name"])
    owner = models.User.get(models.User.email == email)
    bubble = models.Bubble.get(models.Bubble.user == owner,
                               models.Bubble.name == bubble_name)
    if user.id != owner.id and bubble.type != models.Bubble.ENTERPRISE_BUBBLE:
        query = models.BubbleCommiter.select()
        query = query.where(models.BubbleCommiter.user == user,
                            models.BubbleCommiter.bubble == bubble)
        if not query.count():
            flask.abort(404)
    if bubble.type == models.Bubble.ENTERPRISE_BUBBLE:
        status = models.Image.STATUS_PENDING
    else:
        status = models.Image.STATUS_APPROVED
    storage = flask.request.files['photo']
    desc = urllib.unquote_plus(flask.request.form["desc"])
    img = models.Image()
    img.uploaded_by = user
    img.status = status
    img.bubble = bubble
    img.desc = desc
    img.save()
    img.store(storage)
    img.save()
    return {"photo_status": status, "filename": img.name}


@core.api("/change_photo", loggedin=True, cached=30, methods=["POST"])
def change_photo():
    """Chage the data of a given photo of, the user logged in must be a image
     or bubble owner.

    **Logged in:** True

    **Parameters**

        :email:
            User email.
        :bubble_name:
            Bubble name.
        :filename:
            The image filename
        :desc:
            the description of the photo (optional)

    **Return**

        ::

            {
              "action": "change_photo",
              "code": 200,
              "error": false
            }

    """
    user = core.APP_AUTH.get_logged_in_user()
    email = urllib.unquote_plus(flask.request.form["email"])
    bubble_name = urllib.unquote_plus(flask.request.form["bubble_name"])
    filename = urllib.unquote_plus(flask.request.form["filename"])
    owner = models.User.get(models.User.email == email)
    bubble = models.Bubble.get(models.Bubble.user == owner,
                               models.Bubble.name == bubble_name)
    image = models.Image.get(models.Image.bubble == bubble,
                             models.Image.name == filename)
    if owner != user and image.uploaded_by != user:
        return flask.abort(403)
    if "desc" in flask.request.form:
        image.desc = urllib.unquote_plus(flask.request.form["desc"])
    image.save()


@core.api("/give_me_enterprise_bubbles", cached=30)
def give_me_enterprise_bubbles():
    data = {}
    for bubble in models.Bubble.random_enterprise_bubbles():
        data[bubble.name] = bubble.to_dict()
    return {"bubbles": data}


@core.api("/give_me_bubbles", cached=30)
def give_me_bubbles():
    """Returns random listed bubbles.

        **Return**
            {
              "action": "give_me_bubbles",
              "bubbles": {
                "superburbu": {
                  "email": "hello@world.com",
                  "has_password": false,
                  "listed": true,
                  "tags": []
                },
                "test3": {
                  "email": "foo@bar.com",
                  "has_password": false,
                  "listed": true,
                  "tags": []
                },
                "test5": {
                  "email": "foo@bar.com",
                  "has_password": false,
                  "listed": true,
                  "tags": []
                },
                "test6": {
                  "email": "foo@bar.com",
                  "has_password": false,
                  "listed": true,
                  "tags": []
                }
              },
              "code": 200,
              "error": false
            }

    """

    data = {}
    for bubble in models.Bubble.random_bubbles():
        data[bubble.name] = bubble.to_dict()
    return {"bubbles": data}


@core.api("/moderate_bubble", loggedin=True, cached=30,
           transaction=True, enterprise=True, methods=["POST"])
def moderate_bubble():
    """Only enterprise user can moderate only his enterprise bubbles"""
    user = core.APP_AUTH.get_logged_in_user()
    bubble_name = urllib.unquote_plus(flask.request.form["bubble_name"])
    bubble = models.Bubble.get(models.Bubble.name == bubble_name,
                               models.Bubble.user == user,
                               models.Bubble.type == models.Bubble.ENTERPRISE_BUBBLE)
    to_approbe = urllib.unquote_plus(flask.request.form["to_approve"])
    if to_approbe != "null":
        to_approve = to_approbe.split(",")
        query = models.Image.update(status=models.Image.STATUS_APPROVED)
        query = query.where(models.Image.bubble == bubble)
        query = query.where(models.Image.name << to_approve)
        query.execute()
    to_delete = urllib.unquote_plus(flask.request.form["to_delete"])
    if to_delete != "null":
        to_delete = to_delete.split(",")
        query = models.Image.delete().where(models.Image.bubble == bubble)
        query = query.where(models.Image.name << to_delete)
        query.execute()


@core.api("/find/<what>/<list:arguments>", cached=60)
def find(what, arguments):
    """It finds users or bubbles that match the arguments given.

        **Parameters**
            :what:
                ``users`` to look for users. If no match is found the ``users``
                field is returned empty.
                ``bubbles`` to look for bubbles.
                Other keyword raise ``Error 404``
            :argumens[list]:
                Keywords to look for. Several arguments separated by ','.

        **Return**

            for example: the next JSON corresponds with the call /users/Foobar
                {
                  "action": "find",
                  "code": 200,
                  "users": {
                    "hello@world.com": {
                      "lname": "python",
                      "birth": "2012/08/07",
                      "fname": "Foobar",
                      "sex": "m"
                    }
                  },
                  "error": false
                }

    """
    what = urllib.unquote_plus(what)
    found = {}
    for arg in arguments:
        arg = "%{}%".format(arg.lower())
        if what == "users":
            query = models.User.select()
            query = query.where((peewee.fn.Lower(models.User.email) ** arg)
                                | (peewee.fn.Lower(models.User.fname) ** arg)
                                | (peewee.fn.Lower(models.User.lname) ** arg))
            for u in query:
                found[u.email] = u.to_dict()
        elif what == "bubbles":
            query = models.Bubble.select()
            query = query.where((peewee.fn.Lower(models.Bubble.name) ** arg)
                                 | (peewee.fn.Lower(models.Bubble.desc) ** arg))
            query = query.where(models.Bubble.listed == True)
            for b in query:
                found[b.name] = b.to_dict()
        elif what == "ebubbles":
            query = models.Bubble.select()
            query = query.where((peewee.fn.Lower(models.Bubble.name) ** arg)
                                 | (peewee.fn.Lower(models.Bubble.desc) ** arg))
            query = query.where(models.Bubble.listed == True)
            query = query.where(models.Bubble.type ==
                                 models.Bubble.ENTERPRISE_BUBBLE)
            for b in query:
                found[b.name] = b.to_dict()
        else:
            flask.abort(404)
    return {what: found}


@core.api("/invite_to_bubble", loggedin=True,
          transaction=True, methods=["POST"])
def invite_to_bubble():
    user = core.APP_AUTH.get_logged_in_user()
    name = urllib.unquote_plus(flask.request.form["bubble_name"])
    bubble = models.Bubble.get(models.Bubble.user == user,
                               models.Bubble.name == name,
                               models.Bubble.type == models.Bubble.COMMUNITY_BUBBLE)
    mails = set(mail.lower() for mail in
                urllib.unquote_plus(flask.request.form["emails"]).split(",")
                if mail.lower() != user.email.lower()
                and core.RX_MAIL.match(mail))
    for mail in mails:
        to_user = None
        for u in models.User.select().where(models.User.email == mail).limit(1):
            to_user = u
        invitation = models.BubbleInvitation()
        invitation.sended_by = user
        invitation.to_user = to_user
        invitation.email = mail
        invitation.bubble = bubble
        invitation.save()


@core.api("/join_bubble/<email:sender_email>/<bubble_name>/<invitation_hash>",
          loggedin=True, transaction=True)
def join_bubble(sender_email, bubble_name, invitation_hash):
        bubble_name = urllib.unquote_plus(bubble_name)
        user = core.APP_AUTH.get_logged_in_user()
        sended_by = models.User.get(models.User.email == sender_email)
        bubble = models.Bubble.get(models.Bubble.user == sended_by,
                                   models.Bubble.name == bubble_name)
        inv = models.BubbleInvitation.get(models.BubbleInvitation.sended_by == sended_by,
                                          models.BubbleInvitation.bubble == bubble,
                                          models.BubbleInvitation.to_user == user,
                                          models.BubbleInvitation.invitation_hash == invitation_hash)
        if inv.validate_hash(invitation_hash):
            bc = models.BubbleCommiter()
            bc.bubble = bubble
            bc.user = user
            bc.save()
            inv.delete_instance()
        else:
            flask.abort(403)


@core.api("/delete_bubble_invitation", loggedin=True,
          transaction=True, methods=["POST"])
def delete_bubble_invitation():
    user = core.APP_AUTH.get_logged_in_user()
    name = urllib.unquote_plus(flask.request.form["bubble_name"])
    bubble = models.Bubble.get(models.Bubble.user == user,
                               models.Bubble.name == name)
    mails = list(set(mail.lower() for mail in
                urllib.unquote_plus(flask.request.form["emails"]).split(",")
                if mail.lower() != user.email.lower()))
    query = models.BubbleInvitation.select()
    query = query.where(models.BubbleInvitation.bubble == bubble,
                        models.BubbleInvitation.email << mails)
    for inv in query:
        inv.delete_instance()


@core.api("/delete_bubble_commiters", loggedin=True,
           transaction=True, methods=["POST"])
def delete_bubble_commiters():
    user = core.APP_AUTH.get_logged_in_user()
    name = urllib.unquote_plus(flask.request.form["bubble_name"])
    bubble = models.Bubble.get(models.Bubble.user == user, models.Bubble.name == name)
    email = urllib.unquote_plus(flask.request.form["email"])
    cuser = models.User.get(email=email)
    commiter = models.BubbleCommiter.get(models.BubbleCommiter.bubble == bubble,
                                         models.BubbleCommiter.user == cuser)
    commiter.delete_instance()


@core.api("/list_bubble_invitations/<bubble_name>", cached=60, loggedin=True)
def list_bubble_invitations(bubble_name):
    user = core.APP_AUTH.get_logged_in_user()
    bubble_name = urllib.unquote_plus(bubble_name)
    bubble = models.Bubble.get(models.Bubble.user == user,
                               models.Bubble.name == bubble_name)
    invitations = []
    query = models.BubbleInvitation.select()
    query = query.where(models.BubbleInvitation.bubble == bubble)
    for inv in query:
        invitations.append({"email": inv.email,
                            "is_user": inv.to_user != None,
                            "created": core.formatdate(inv.creation_datetime),
                            "sended": inv.sended,
                            "sended_by": user.email,
                            "bubble_name": bubble.name,
                            "has_password": bubble.has_password,
                            "desc": bubble.desc,
                            "listed": inv.bubble.listed})
    return {"invitations": invitations}


@core.api("/list_bubble_commiters/<bubble_name>", cached=60, loggedin=True)
def list_bubble_commiters(bubble_name):
    user = core.APP_AUTH.get_logged_in_user()
    bubble_name = urllib.unquote_plus(bubble_name)
    bubble = models.Bubble.get(models.Bubble.user == user,
                               models.Bubble.name == bubble_name)
    cis = []
    query = models.BubbleCommiter.select()
    query = query.where(models.BubbleCommiter.bubble == bubble)
    for ci in query:
        cis.append({"email": ci.user.email,
                    "joined": core.formatdate(ci.joined_datetime)})
    return {"commiters": cis}


@core.api("/my_bubbles_invitations", cached=60, loggedin=True)
def my_bubbles_invitations():
    user = core.APP_AUTH.get_logged_in_user()
    invitations = []
    query = models.BubbleInvitation.select()
    query = query.where(models.BubbleInvitation.to_user == user)
    for inv in query:
        invitations.append({"email": inv.email,
                            "is_user": inv.to_user != None,
                            "created": core.formatdate(inv.creation_datetime),
                            "sended": inv.sended,
                            "sended_by": inv.sended_by.email,
                            "bubble_name": inv.bubble.name,
                            "has_password": inv.bubble.has_password,
                            "listed": inv.bubble.listed})
        return {"invitations": invitations}


@core.api("/set_enterprise_user", cached=60, admin=True, methods=["POST"])
def set_enterprise_user():
    email = urllib.unquote_plus(flask.request.form["email"])
    is_enterprise = urllib.unquote_plus(flask.request.form["is_enterprise"])
    user = models.User.get(models.User.email == email)
    user.is_enterprise = (is_enterprise == "true")
    user.save()


#===============================================================================
# MAIN
#===============================================================================

if __name__ == "__main__":
    print(__doc__)
