#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2012 Liricus SRL <info@liricus.com.ar>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Fra


#===============================================================================
# DOCS
#===============================================================================

"""Loads the Flask application for the socialframe server.

Instances the mail client, database and authentication client.

"""


#===============================================================================
# IMPORTS
#===============================================================================

import sys
import traceback
import functools
import datetime
import re
import time
import urllib

import flask
import flask_mail
from flaskext import uploads
from flask.ext import cache

from flask_peewee import db

from werkzeug import routing

from socialframe_server import settings


#===============================================================================
# CONSTANTS
#===============================================================================

APP = flask.Flask("socialframe_server")
APP.config.from_object(settings)

APP_DB = db.Database(APP)

APP_MAIL = flask_mail.Mail(APP)

APP_CACHE = cache.Cache(APP)

APP_AUTH = None # this is setted when models is imported

RX_MAIL = re.compile("^[a-zA-Z0-9._%-]+@[a-zA-Z0-9._%-]+.[a-zA-Z]{2,6}$")

FILTER2EXT = {
    "JPEG": ("jpg", "jpeg", "jpe", "jiff"),
    "PNG": ("png",)
}

EXT2FILTER = {}
for k, vs in FILTER2EXT.items():
    for v in vs:
        EXT2FILTER[v] = k

VALID_IMAGES = EXT2FILTER.keys()


#===============================================================================
# VALIDATORS
#===============================================================================

errors = []
if "original" in settings.IMAGE_SIZES:
    errors.append("'original' image size name is not allowed")

if errors:
    msg = "\n".join(errors)
    raise ValueError(msg)


#===============================================================================
# CONFIGS
#===============================================================================

if settings.MAX_UPLOAD_SIZE > 0:
    uploads.patch_request_class(APP, settings.MAX_UPLOAD_SIZE)


#===============================================================================
# CONVERTERS
#===============================================================================

class ListConverter(routing.BaseConverter):
    def to_python(self, value):
        value = urllib.unquote_plus(value)
        l = []
        for v in value.split(','):
            if v == "null":
                v = None
            l.append(v)
        return l
    def to_url(self, values):
        l = []
        for v in values:
            if v is None:
                v = "null"
            l.append(routing.BaseConverter.to_url(v))
        return ','.join(l)
APP.url_map.converters['list'] = ListConverter


class NullConverter(routing.BaseConverter):
    def to_python(self, value):
        value = urllib.unquote_plus(value)
        return None if value == "null" else value
    def to_url(self, value):
        return routing.BaseConverter.to_url(
            "null" if value == None else value
        )
APP.url_map.converters['null'] = NullConverter


class EmailConverter(routing.BaseConverter):
    def to_python(self, value):
        value = urllib.unquote_plus(value)
        if not RX_MAIL.match(value):
            flask.abort(400)
        return value
    def to_url(self, value):
        return value
APP.url_map.converters['email'] = EmailConverter



#===============================================================================
# ERROR HANDLERS
#===============================================================================

@APP.errorhandler(404)
@APP.errorhandler(403)
@APP.errorhandler(500)
@APP.errorhandler(400)
@APP.errorhandler(Exception)
def errorhandler_json(ex):
    """Returns error code and information in JSON format.

        **Errors handled**

        The Errors are ``400``, ``403``, ``404``, ``500``.

    """

    code = getattr(ex, "code", None)
    if isinstance(code, basestring) and code.isdigit():
        code = int(code)
    elif not isinstance(code, int):
        code = 500

    data = {
        "error": True,
        "code": code,
        "msg": ex.message,
    }
    if settings.DEBUG:
        traceinfo = "".join(traceback.format_exception(*sys.exc_info()))
        data["traceback"] = traceinfo
    return json_response(**data)


#===============================================================================
# FUNCTIONS
#===============================================================================

def absurl(*args):
    """Creates an absolute url mapped with the given arguments and returns it.

    """

    parts = [settings.FRONT_END_URL]
    for arg in args:
        if arg.startswith("/"):
            arg = arg[1:]
        parts.append(arg)
    return "/".join(parts)


def json_response(**args):
    """Creates a default json response for socialframe

    """
    data = dict(args)
    data["action"] = flask.request.endpoint
    if "error" not in data:
        data["error"] = False
    data["code"] = data.get("code", 200)
    resp = flask.jsonify(**data)
    resp.status_code = data["code"]
    return resp
json_data = json_response


_unique = int(time.time())
def unique():
    """Return a unique code for a statting with sistem time when
    the project is started

    """
    global _unique
    _unique += 1
    return _unique


_jobs = set()
def run_jobs():
    global _jobs
    for func in _jobs:
        print "Job '{}' run...".format(func.__name__)
        with APP_DB.database.transaction():
            func()


def parsedate(ds):
    return datetime.datetime.strptime(ds, "%Y%m%d")


def formatdate(dt):
    return dt.strftime("%Y%m%d")


def formatdatetime(dt):
    return dt.strftime("%Y%m%d_%H:%M:%s")


#===============================================================================
# DECORATORS
#===============================================================================

def socialframe_jsonify(func):
    """Return the data dict as default socialframe json response

    """
    @functools.wraps(func)
    def _wrap(*args, **kwargs):
        data = func(*args, **kwargs) or {}
        return json_response(**data)
    return _wrap


def login_required(func):
    """Checks if user is logged, then original function ``func`` proceeds.
    If not throws ``Error 403``.

    """

    @functools.wraps(func)
    def _wrap(*args, **kwargs):
        if APP_AUTH.get_logged_in_user():
            return func(*args, **kwargs)
        else:
            flask.abort(403)

    return _wrap


def transactional(func):
    """Execute the given function as transactional

    """
    @functools.wraps(func)
    def _wrap(*args, **kwargs):
        with APP_DB.database.transaction():
            return func(*args, **kwargs)
    return _wrap


def job(func):
    """Register a function as a backend job to run with  run_jobs

    """
    global _jobs

    @functools.wraps(func)
    def _wrap(*args, **kwargs):
        with APP.app_context():
            return func(*args, **kwargs)

    _jobs.add(_wrap)
    return _wrap


def is_admin(func):
    """Execute the call only if the  user ``is_superuser`` is ``True``"""

    @functools.wraps(func)
    def _wrap(*args, **kwargs):
        user = APP_AUTH.get_logged_in_user()
        if user and user.is_superuser:
            return func(*args, **kwargs)
        else:
            flask.abort(403)

    return _wrap


def is_enterprise(func):
    """Execute the call only if the  user ``is_enterprise`` is ``True``"""

    @functools.wraps(func)
    def _wrap(*args, **kwargs):
        user = APP_AUTH.get_logged_in_user()
        if user and user.is_enterprise:
            return func(*args, **kwargs)
        else:
            flask.abort(403)

    return _wrap


def api(path, cached=0.0, json=True, admin=False,
        loggedin=False, transaction=False, enterprise=False,
        *rargs, **rkwargs):
    """This is the default meta decorator for socialframe

    """

    def _wrap(func):
        if admin:
            func = is_admin(func)
        if enterprise:
            func = is_enterprise(func)
        if transaction:
            func = transactional(func)
        if loggedin:
            func = login_required(func)
        if json:
            func = socialframe_jsonify(func)
        if cached:
            func = APP_CACHE.memoize(timeout=cached)(func)
        func = APP.route(path, *rargs, **rkwargs)(func)
        return func

    return _wrap


#===============================================================================
# MAIN
#===============================================================================

if __name__ == "__main__":
    print(__doc__)
